
#include "db.h"
#include "httpd.h"
#include "socket.h"
#include "malloc.h"
#include "timer.h"

#include <ctype.h>

static int max_persist_requests = 32;	// ʐMł̍ő僊NGXg
void httpd_set_max_persist_requests( int i ){ max_persist_requests = i; }

static int request_timeout[] = { 2500, 60*1000 };	// ^CAEg(ŏA)
void httpd_set_request_timeout( int idx, int time ){ request_timeout[idx]=time; }

static char document_root[256]="./httpd/";	// hLg[g
void httpd_set_document_root( const char *str ){ strcpy( document_root,str ); }

enum HTTPD_STATUS {
	HTTPD_REQUEST_WAIT = 0,		// NGXg҂
	HTTPD_REQUEST_WAIT_POST,	// NGXg҂(post)
	HTTPD_REQUEST_OK,			// NGXgߊ
	HTTPD_SEND_HEADER,			// wb_M
	HTTPD_WAITING_SEND			// f[^MI܂ő҂Ă
};

struct httpd_session_data {
	int fd;
	int status;
	int http_ver;
	int header_len;
	int data_len;
	int method;
	int persist;
	int request_count;
	unsigned int tick;
	const unsigned char* url;
	const unsigned char* query;
};

int httpd_strcasencmp(const char *s1, const char *s2,int len) {
	while(len-- && (*s1 || *s2) ) {
		if((*s1 | 0x20) != (*s2 | 0x20)) {
			return ((*s1 | 0x20) > (*s2 | 0x20) ? 1 : -1);
		}
		s1++; s2++;
	}
	return 0;
}

// httpd ɓĂy[WƁAĂяoR[obN֐̈ꗗ
struct dbt* httpd_files;

static int httpd_db_final(void *key,void *data,va_list ap)
{
	char *url = key;

	free(url);

	return 0;
}

static void do_final_httpd(void) {
	db_final(httpd_files,httpd_db_final);
}

void do_init_httpd(void) {
	httpd_files = strdb_init(0);
	atexit(do_final_httpd);
}

// httpd Ƀy[Wǉ
// for ȂǂŃy[Wł悤ɁAkey strdup()̂g

void httpd_pages(const char* url,void (*httpd_func)(struct httpd_session_data*,const char*)) {
	if(strdb_search(httpd_files,url+1) == NULL) {
		strdb_insert(httpd_files,strdup(url+1),httpd_func);
	} else {
		strdb_insert(httpd_files,url+1,httpd_func);
	}
}

static void(*httpd_default)(struct httpd_session_data* sd,const char* url);

void httpd_default_page(void(*httpd_func)(struct httpd_session_data* sd,const char* url)) {
	httpd_default = httpd_func;
}

const char *httpd_get_error( struct httpd_session_data* sd, int* status )
{
	const char* msg;
	// httpd ̃Xe[^X߂
	switch(*status) {
	case 200: msg = "OK";           break;
	case 400: msg = "Bad Request";  break;
	case 401: msg = "Unauthorized"; break;	// gp
	case 403: msg = "Forbidden";    break;	// gp
	case 404: msg = "Not Found";    break;
	case 408: msg = "Request Timedout"; break;
	case 411: msg = "Length Required"; break;
	case 413: msg = "Request Entity Too Large"; break;
	default:
		*status = 500; msg = "Internal Server Error"; break;
	}
	return msg;
}

void httpd_send_error(struct httpd_session_data* sd,int status) {
	const char* msg = httpd_get_error( sd, &status );
	httpd_send(sd,status,"text/plain",strlen(msg),msg);
}

void httpd_send_head(struct httpd_session_data* sd,int status,const char *content_type,int content_len) {
	char head[256];
	int  len;
	const char* msg;

	if(sd->status != HTTPD_REQUEST_OK) return;
	msg = httpd_get_error( sd, &status );
	
	if(content_len == -1) {
		len = sprintf(
			head,
			"HTTP/1.%d %d %s\r\nContent-Type: %s\r\n\r\n",
			sd->http_ver,status,msg,content_type
		);
		sd->persist = 0; // Ȃ̂ŁAؒfɂ
	} else {
		len = sprintf(
			head,
			"HTTP/1.%d %d %s\r\nContent-Type: %s\r\nContent-Length: %d\r\n\r\n",
			sd->http_ver,status,msg,content_type,content_len
		);
	}
	memcpy(WFIFOP(sd->fd,0),head,len);
	WFIFOSET(sd->fd,len);
	sd->status   = HTTPD_SEND_HEADER;
	sd->data_len = content_len;
}

void httpd_send_data(struct httpd_session_data* sd,int content_len,const void *data) {
	const char* msg = (const char*)data;
	if(sd->status == HTTPD_REQUEST_OK) {
		// wb_̑MYĂ̂ŁAKɕ₤
		httpd_send_head(sd,200,"application/octet-stream",-1);
	} else if(sd->status != HTTPD_SEND_HEADER && sd->status != HTTPD_WAITING_SEND) {
		return;
	}
	sd->data_len -= content_len;

	// ȃTCỸt@CMo悤ɕđ
	while(content_len > 0) {
		int send_byte = content_len;
		if(send_byte > 12*1024) send_byte = 12*1024;
		memcpy(WFIFOP(sd->fd,0),msg,send_byte);
		WFIFOSET(sd->fd,send_byte);
		msg += send_byte; content_len -= send_byte;
	}
	sd->status = HTTPD_WAITING_SEND;
}

void httpd_send(struct httpd_session_data* sd,int status,const char *content_type,int content_len,const void *data) {
	httpd_send_head(sd,status,content_type,content_len);
	httpd_send_data(sd,content_len,data);
}

int httpd_disconnect(int fd) {
	// nothing to do
	return 0;
}

void httpd_parse_header(struct httpd_session_data* sd);
void httpd_parse_request_ok(struct httpd_session_data *sd);

int  httpd_parse(int fd) {
	struct httpd_session_data *sd = (struct httpd_session_data*)session[fd]->session_data2;
	if(sd == NULL) {
		sd = (struct httpd_session_data*)aCalloc(sizeof(struct httpd_session_data),1);
		sd->fd  = fd;
		session[fd]->func_destruct = httpd_disconnect;
		session[fd]->session_data2  = sd;
		sd->tick = gettick();
		sd->persist = 0;
		sd->request_count = 0;
	}
	switch(sd->status) {
	case HTTPD_REQUEST_WAIT:
		// NGXg҂
		if(RFIFOREST(fd) > 1024) {
			// NGXĝŁAG[
			sd->status = HTTPD_REQUEST_OK;
			httpd_send_error(sd,400); // Bad Request
		} else if( (int)( gettick() - sd->tick ) > request_timeout[sd->persist] ) {
			// NGXgɎԂ肷Ă̂ŁAG[
			sd->status = HTTPD_REQUEST_OK;
			httpd_send_error(sd,408); // Request Timeout
		} else if(sd->header_len == RFIFOREST(fd)) {
			// ԂȑOƓȂ̂ŁANGXgĉ͂Kv͖
		} else {
			int limit = RFIFOREST(fd);
			unsigned char *req = RFIFOP(fd,0);
			sd->header_len = RFIFOREST(fd);
			do {
				if(*req == '\n' && limit > 0) {
					limit--; req++;
					if(*req == '\r' && limit > 0) { limit--; req++; }
					if(*req == '\n') {
						// HTTPwb_̏I_
						*req   = 0;
						sd->header_len = (req - RFIFOP(fd,0)) + 1;
						httpd_parse_header(sd);
						break;
					}
				}
			} while(req++,--limit > 0);
		}
		break;
	case HTTPD_REQUEST_WAIT_POST:
		if(RFIFOREST(sd->fd) >= sd->header_len) {
			unsigned char temp = RFIFOB(sd->fd,sd->header_len);
			RFIFOB(sd->fd,sd->header_len) = 0;
			httpd_parse_request_ok(sd);
			RFIFOB(sd->fd,sd->header_len) = temp;
		}
		break;
	case HTTPD_REQUEST_OK:
	case HTTPD_SEND_HEADER:
		// NGXgI܂܉MĂȂԂȂ̂ŁA
		// ؒf
		session[fd]->eof = 1;
		break;
	case HTTPD_WAITING_SEND:
		// f[^̑MI܂őҋ@
		if(session[fd]->wdata_size == session[fd]->wdata_pos) {
			// HTTP/1.0͎蓮ؒf
//			if(sd->http_ver == 0) {
			if(sd->persist == 0 ||  ++sd->request_count >= max_persist_requests ) {
				session[fd]->eof = 1;
			}
			// RFIFO 烊NGXgf[^̏ƍ\̂̏
			RFIFOSKIP(fd,sd->header_len);
			sd->status     = HTTPD_REQUEST_WAIT;
			sd->tick       = gettick();
			sd->header_len = 0;
			sd->query      = NULL;
//			sd->http_ver   = 0;	// ver ͕ێ
			sd->method     = HTTPD_METHOD_UNKNOWN;
			// printf("httpd_parse: [% 3d] request sended RFIFOREST:%d\n",fd,RFIFOREST(fd));
		}
		break;
	}
	return 0;
}

void httpd_parse_header_sub( struct httpd_session_data* sd, const char *p1, int* plen )
{
	int len = 0;
	// HTTP̃o[W𒲍
	if(!strncmp(p1 ,"HTTP/1.1",8)) {
		sd->http_ver = 1;
		sd->persist  = 1;
	} else {
		sd->http_ver = 0;
		sd->persist  = 0;
	}

	p1  = strchr(p1,'\n');
	while(p1) {
		// Content-Length: ̒
		if(!httpd_strcasencmp(p1+1,"Content-Length: ",16)) {
			len = atoi(p1 + 17);
		}
		// Connection: ̒
		if(!httpd_strcasencmp(p1+1,"Connection: ",12)) {
			if( httpd_strcasencmp(p1+13,"close",5)==0 && sd->http_ver==1 )
				sd->persist = 0;
			if( httpd_strcasencmp(p1+13,"Keep-Alive",10)==0 && sd->http_ver==0 )
				sd->persist = 1;
		}
		p1 = strchr(p1+1,'\n');
	}
	if(plen) *plen = len;
	return;
}

void httpd_parse_header(struct httpd_session_data* sd) {
	int  i;
	int  status = 400; // Bad Request
	unsigned char* req    = RFIFOP(sd->fd,0);
	do {
		if(!strncmp(req,"GET /",5)) {
			// GET NGXg
			req += 5;
			for(i = 0;req[i]; i++) {
				if(req[i] == ' ' || req[i] == '?') break;
				if(!isalnum(req[i]) && req[i] != '.' && req[i] != '_' && req[i] != '-') break;
			}
			if(req[i] == ' ') {
				req[i]     = 0;
				sd->url    = req;
				sd->query  = NULL;
				sd->status = HTTPD_REQUEST_OK;
			} else if(req[i] == '?') {
				req[i]    = 0;
				sd->query = &req[++i];
				for(;req[i];i++) {
					if(
						isalnum(req[i]) || req[i] == '+' || req[i] == '%' || req[i] == '&' ||
						req[i] == '='
					) {
						continue;
					} else {
						break;
					}
				}
				if(req[i] != ' ') {
					break;
				}
				req[i]     = 0;
				sd->url    = req;
			} else {
				break;
			}
			// wb_
			httpd_parse_header_sub( sd, &req[i+1], NULL );
			
			// printf("httpd: request %s %s\n",sd->url,sd->query);
			sd->method     = HTTPD_METHOD_GET;
			httpd_parse_request_ok(sd);
		} else if(!strncmp(req,"POST /",6)) {
			int  len;
			req += 6; status = 404;
			for(i = 0;req[i]; i++) {
				if(req[i] == ' ') break;
				if(!isalnum(req[i]) && req[i] != '.' && req[i] != '_' && req[i] != '-') break;
			}
			if(req[i] != ' ') {
				break;
			}
			req[i]     = 0;
			sd->url    = req;

			// wb_
			httpd_parse_header_sub( sd, &req[i+1], &len );
			
			if(len <= 0 || len >= 32*1024) {
				// Ƃ肠32KBȏ̃NGXg͕s
				status = ( len==0 )? 411 : ( len>32*1024 )? 413 : 400;
				break;
			}
			
			sd->query      =  RFIFOP(sd->fd,sd->header_len);
			sd->method     = HTTPD_METHOD_POST;
			sd->header_len += len;
			if(RFIFOREST(sd->fd) >= sd->header_len) {
				unsigned char temp = RFIFOB(sd->fd,sd->header_len);
				RFIFOB(sd->fd,sd->header_len) = 0;
				httpd_parse_request_ok(sd);
				RFIFOB(sd->fd,sd->header_len) = temp;
			} else {
				// POST̃f[^Ă̂҂
				sd->status = HTTPD_REQUEST_WAIT_POST;
			}
		} else {
			break;
		}
	} while(0);
	if(sd->status == HTTPD_REQUEST_WAIT) {
		sd->status = HTTPD_REQUEST_OK;
		httpd_send_error(sd,status); 
	}
}

void httpd_parse_request_ok(struct httpd_session_data *sd) {
	void (*httpd_parse_func)(struct httpd_session_data*,const char*);
	sd->status = HTTPD_REQUEST_OK;

	// t@C܂̂ŁAy[W
	// printf("httpd_parse: [% 3d] request /%s\n",fd,req);
	httpd_parse_func = strdb_search(httpd_files,sd->url);
	if(httpd_parse_func == NULL) {
		httpd_parse_func = httpd_default;
	}
	if(httpd_parse_func == NULL) {
		httpd_send_error(sd,404); // Not Found
	} else {
		httpd_parse_func(sd,sd->url);
		if(sd->status == HTTPD_REQUEST_OK) {
			httpd_send_error(sd,404); // Not Found
		}
	}
	if(sd->http_ver == 1 && sd->data_len) {
		// ςȃf[^(Ȃ̑Ȃc)
		printf("httpd_parse: send size mismatch when parsing /%s\n",sd->url);
		session[sd->fd]->eof = 1;
	}
	if(sd->status == HTTPD_REQUEST_OK) {
		httpd_send_error(sd,404); 
	}
}

char* httpd_get_value(struct httpd_session_data* sd,const char* val) {
	int src_len = strlen(val);
	const unsigned char* src_p = sd->query;
	if(src_p == NULL) return aStrdup("");

	do {
		if(!memcmp(src_p,val,src_len) && src_p[src_len] == '=') {
			break;
		}
		src_p = strchr(src_p + 1,'&');
		if(src_p) src_p++;
	} while(src_p);

	if(src_p != NULL) {
		// ړI̕
		const unsigned char* p2;
		int   dest_len;
		char* dest_p;
		src_p += src_len + 1;
		p2     = strchr(src_p,'&');
		if(p2 == NULL) {
			src_len = strlen(src_p);
		} else {
			src_len = (p2 - src_p);
		}
		dest_p   = aMalloc(src_len + 1);
		dest_len = 0;
		while(src_len > 0) {
			if(*src_p == '%' && src_len > 2) {
				int c1 = 0,c2 = 0;
				if(src_p[1] >= '0' && src_p[1] <= '9') c1 = src_p[1] - '0';
				if(src_p[1] >= 'A' && src_p[1] <= 'F') c1 = src_p[1] - 'A' + 10;
				if(src_p[1] >= 'a' && src_p[1] <= 'f') c1 = src_p[1] - 'a' + 10;
				if(src_p[2] >= '0' && src_p[2] <= '9') c2 = src_p[2] - '0';
				if(src_p[2] >= 'A' && src_p[2] <= 'F') c2 = src_p[2] - 'A' + 10;
				if(src_p[2] >= 'a' && src_p[2] <= 'f') c2 = src_p[2] - 'a' + 10;
				dest_p[dest_len++] = (c1 << 4) | c2;
				src_p += 3; src_len -= 3;
			} else if(*src_p == '+') {
				dest_p[dest_len++] = ' ';
				src_p++; src_len--;
			} else {
				dest_p[dest_len++] = *(src_p++); src_len--;
			}
		}
		dest_p[dest_len] = 0;
		return dest_p;
	}
	return aStrdup("");
}

int httpd_get_method(struct httpd_session_data* sd) {
	return sd->method;
}

unsigned int httpd_get_ip(struct httpd_session_data *sd) {
	return *(unsigned int*)(&session[sd->fd]->client_addr.sin_addr);
}

// MIME^CvBvȂ̂肵āAcapplication/octet-stream
static const char* httpd_mimetype(const char* url) {
	char *ext = strrchr(url,'.');
	if(ext) {
		if(!strcmp(ext,".html")) return "text/html";
		if(!strcmp(ext,".htm"))  return "text/html";
		if(!strcmp(ext,".css"))  return "text/css";
		if(!strcmp(ext,".js"))   return "text/javascript";
		if(!strcmp(ext,".txt"))  return "text/plain";
		if(!strcmp(ext,".gif"))  return "image/gif";
		if(!strcmp(ext,".jpg"))  return "image/jpeg";
		if(!strcmp(ext,".jpeg")) return "image/jpeg";
		if(!strcmp(ext,".png"))  return "image/png";
		if(!strcmp(ext,".xbm"))  return "image/xbm";
		if(!strcmp(ext,".zip"))  return "application/zip";
	}
	return "application/octet-stream";
}

void httpd_send_file(struct httpd_session_data* sd,const char* url) {
	FILE *fp;
	int  file_size;
	char file_buf[8192];
	if(sd->status != HTTPD_REQUEST_OK) return;
	if(url[0] == '\0') url = "index.html";
	
	// url ̍ő咷͖1010oCgȂ̂ŁAobt@I[o[t[̐Sz͖
	sprintf(file_buf,"%s%s",document_root,url);

	fp = fopen(file_buf,"rb");
	if(fp == NULL) {
		httpd_send_error(sd,404);
	} else {
		fseek(fp,0,SEEK_END);
		file_size = ftell(fp);
		fseek(fp,0,SEEK_SET);
		httpd_send_head(sd,200,httpd_mimetype(url),file_size);
		while(file_size > 0) {
			int read_byte = file_size;
			if(file_size > 8192) read_byte = 8192;
			fread(file_buf,1,read_byte,fp);
			httpd_send_data(sd,read_byte,file_buf);
			file_size -= read_byte;
		}
		fclose(fp);
	}
}
